/*
 * Weller Copyright (c) 2010, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/event.h>
#include <sys/time.h>

#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <errno.h>

#define VERSION "1.11"

#define LOCKBUFSIZE 131072
#define XOR_CHAR 129

char *host = "", *port = "7331", buffer[ 1024 ],
   *lock_msg = "<display locked>\n", *ulock_msg = "<display unlocked>\n",
   *filename = NULL, lock_buffer[ LOCKBUFSIZE ], *lock_ptr = lock_buffer;

int fd = -1, lock = 0, lock_count = 0, lock_free = LOCKBUFSIZE,
   lock_len = 17, ulock_len = 19;

FILE *transcript = NULL;

void get_args( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "t:" )) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 't':
            filename = optarg;
            break;
      }

   argc -= optind;
   argv += optind;

   if ( argc < 1 )
   {
      fprintf( stderr, "hostname not specified\n" );
      exit( 1 );
   }

   host = argv[ 0 ];

   if ( argc > 1 )
      port = argv[ 1 ];

   if ( filename != NULL && 
        ( transcript = fopen( filename, "w+" )) == NULL )
   {
      fprintf( stderr, "cannot open transcript file: %s", strerror( errno ));
      exit( 1 );
   }
}

void open_remote()
{
   int result;
   struct addrinfo hints, *res, *ptr;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_family = PF_UNSPEC;
   hints.ai_socktype = SOCK_STREAM;

   if (( result = getaddrinfo( host, port, &hints, &res )))
   {
      fprintf( stderr, "getaddrinfo(): %s\n", ( char *)gai_strerror( result ));
      exit( 1 );
   }

   for( ptr = res; ptr != NULL; ptr = ptr->ai_next )
   {
      fd = socket( ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol );

      if ( fd == -1 )
         continue;

      if ( connect( fd, ptr->ai_addr, ptr->ai_addrlen ) == -1 )
      {
         close( fd );
         fd = -1;
         continue;
      }

      break;
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( fd == -1 )
   {
      fprintf( stderr, "socket()/connect(): %s\n", strerror( errno ));
      exit( 1 );
   }
}

void buffer_incoming( int count )
{
   if ( count > lock_free )
      return;

   bcopy( buffer, lock_ptr, count );

   lock_ptr += count;
   lock_count += count;
   lock_free -= count;
}

void consume_buffer()
{
   write( 1, lock_buffer, lock_count );

   if ( transcript != NULL && 
        ! fwrite( lock_buffer, lock_count, 1, transcript ))
      fprintf( stderr, "fwrite( transcript ): %s", strerror( errno ));

   lock_ptr = lock_buffer;
   lock_count = 0;
   lock_free = LOCKBUFSIZE;
}

void xor( int count )
{
   char *ptr;

   for( ptr = buffer; count; --count, ++ptr )
      if (( unsigned char )*ptr != XOR_CHAR )
         *ptr ^= XOR_CHAR;
}

void read_from_client()
{
   int count;

   if (( count = read( 1, buffer, sizeof( buffer ))) < 0 )
   {
      fprintf( stderr, "read( <client> ): %s", strerror( errno ));
      exit( 1 );
   }

   if ( !count )
      exit( 0 );

   if ( *buffer == '\r' || *buffer == '\n' )
   {
      if ( lock ^= 1 )
      {
         write( 1, lock_msg, lock_len );

         if ( transcript != NULL && fputs( lock_msg, transcript ) == EOF )
            fprintf( stderr, "fputs( transcript ): %s", strerror( errno ));
      }
      else
      {
         write( 1, ulock_msg, ulock_len );

         if ( transcript != NULL && fputs( ulock_msg, transcript ) == EOF )
            fprintf( stderr, "fputs( transcript ): %s", strerror( errno ));

         consume_buffer();
      }

      return;
   }

   if ( transcript != NULL && ! fwrite( buffer, count, 1, transcript ))
      fprintf( stderr, "fwrite( transcript ): %s", strerror( errno ));

   xor( count );

   if ( write( fd, buffer, count ) < 0 )
   {
      fprintf( stderr, "write( <server> ): %s", strerror( errno ));
      exit( 1 );
   }
}

void read_from_server()
{
   int count;

   if (( count = read( fd, buffer, sizeof( buffer ))) < 0 )
   {
      fprintf( stderr, "read( <server> ): %s\n", strerror( errno ));
      exit( 1 );
   }

   if ( !count )
   {
      fprintf( stderr, "pickwick closed connection\n" );
      exit( 0 );
   }

   xor( count );

   if ( lock )
      buffer_incoming( count );
   else 
   {
      if ( write( 1, buffer, count ) < 0 )
      {
         fprintf( stderr, "write( <client> ): %s\n", strerror( errno ));
         exit( 1 );
      }
   
      if ( transcript != NULL && ! fwrite( buffer, count, 1, transcript ))
         fprintf( stderr, "fwrite( transcript ): %s", strerror( errno ));
   }
}
   
void ev_set( int desc, short filter, u_short flags, struct kevent *inqueue )
{
   inqueue->ident = desc;
   inqueue->filter = filter;
   inqueue->fflags = 0;
   inqueue->flags = flags;
   inqueue->udata = NULL;
}

void interact()
{
   int kq, n, out, inq;
   struct kevent outqueue[ 2 ];
   struct kevent inqueue[ 2 ], *qptr;

   *lock_buffer = '\0';

   if (( kq = kqueue()) < 0 )
   {
      fprintf( stderr, "kqueue(): %s", strerror( errno ));
      exit( 1 );
   }

   ev_set( fd, EVFILT_READ, EV_ADD | EV_ENABLE, &inqueue[ 0 ] );
   ev_set( fileno( stdin ), EVFILT_READ, EV_ADD | EV_ENABLE, &inqueue[ 1 ] );

   qptr = inqueue;
   inq = 2;

   for( ; ; )
   {
      out = kevent( kq, qptr, inq, outqueue, 2, NULL );

      qptr = NULL;
      inq = 0;

      if ( out <= 0 )
         continue;

      for( n = 0; n < out; ++n )
      {
         if ( outqueue[ n ].ident < 0 || outqueue[ n ].flags & EV_ERROR )
            continue;

         if ( outqueue[ n ].ident == fd )
            read_from_server();
         else
            read_from_client();
      }
   }
}

int main( int argc, char **argv )
{
   get_args( argc, argv );

   open_remote();
   interact();

   return 0;
}
